#include <EEPROM.h>
#include <PID_Beta6.h>
#include <adc.h>
#include <display.h>
#include <fet.h>
#include <keypad.h>
#include <pressure.h>
#include <servo.h>
#include <temp.h>
#include <timer.h>
#include <ui.h>
#include <util.h>
#include <avr/io.h>    

#define ALARM_FET FET6
#define EJECTOR_FET FET7

#define GRATE_OFF 0
#define GRATE_ON 1
#define GRATE_PRATIO 2

//Datalogging variables
#include "WProgram.h"
void setup();
void loop();
void DoAlarmUpdate();
void DoAlarm();
void DoDatalogging();
void Ejector_Init();
void DoEjector();
void DoFlow();
double CalcFlowmeter(double CfA0, double dP, double density);
void DoGrate();
void DoPressure();
void CalibratePressureSensors();
void LoadPressureSensorCalibration();
void UpdateCalibratedPressure();
float smooth(int data, float filterVal, float smoothedVal);
void DoSerialIn();
void PrintEjectorUpdate(double P, double I, double D, double nP, double nI, double nD);
void PulseServo(int servo_pin,double angle);
void DoServos();
int lineCount = 0;

// Grate turning variables
int GRATE_SOLENOID = FET5;
//int GRATE_MOTOR_0 = FET7;
//int GRATE_MOTOR_1 = FET6;
unsigned long gratePeriod = 300000;
int gratePulseLength = 30000;
int grateMode = GRATE_PRATIO;
unsigned long nextGrate;
float pRatio, pRatioSmooth;
boolean pRatioHigh;
boolean grateOn;

// Flow variables
float CfA0_air_rct =0.42123;
float CfA0_gas = 0.81046;
double air_rct_flow;
double gas_flow;

int loopPeriod0 = 5000;
unsigned long nextTime0;
int loopPeriod1 = 1000;
unsigned long nextTime1;
int loopPeriod2 = 100;
unsigned long nextTime2;
int loopPeriod3 = 10;
unsigned long nextTime3;

//Ejector variables
double ejector_setpoint, ejector_input, ejector_output;
PID ejector_PID(&ejector_input, &ejector_output, &ejector_setpoint,0,0,0);
int ejector_period = 1000;
unsigned long ejector_nextCompute = 0;
double ejector_P = -15;
double ejector_I = 10;
double ejector_D = 0;

//Hertz
int hertz = 0;

// Pressure variables
int Press_Calib[6];
int Press[6]; //values corrected for sensor offset (calibration)
int P_air_in;
int P_gas_out;
int P_comb;
float P_comb_smooth;
int P_reactor;

//Servo
int servo_alt = 0; //used to pulse every other time through loop (~20 ms)

//Servo0
float servo0_pos = 0;
float servo0_db = 0; // used to deadband the servo movement

//Servo1
float servo1_pos;
float servo1_db = 0; // used to deadband the servo movement

// Alarm
int auger_on_length = 0;
int auger_off_length = 0;
unsigned int auger_on_alarm_point = 300;
unsigned int auger_off_alarm_point = 900;
boolean alarm;
int LOW_FUEL_TC = 3;
int alarm_interval = 5; // in seconds

void setup() {
  GCU_Setup(V3,FULLFILL,P777222);
  //
  DDRJ |= 0x80;      
  PORTJ |= 0x80;   
  delay(1);	
  
  // timer initialization
  nextTime0 = millis() + loopPeriod0;
  nextTime1 = millis() + loopPeriod1;
  nextTime2 = millis() + loopPeriod2;
  nextTime3 = millis() + loopPeriod3;

  Ejector_Init(); // params should be pulled out to here
  LoadPressureSensorCalibration();
  //LoadLambda(); - must save lambda data first?
  Serial.begin(57600);

  Disp_Init();
  Kpd_Init();
  UI_Init();
  ADC_Init();
  Temp_Init();
  Press_Init();
  Fet_Init();
  Servo_Init();
  Timer_Init();

  Disp_Reset();
  Kpd_Reset();
  UI_Reset();
  ADC_Reset();
  Temp_Reset();
  Press_Reset();
  Fet_Reset();
  Servo_Reset();
  Timer_Reset();
}

void loop() {
  int key;
  if (millis() >= nextTime3) {
    nextTime3 += loopPeriod3;
    // first, read all KS's sensors
    Temp_ReadAll();  // reads into array Temp_Data[], in deg C
    Press_ReadAll(); // reads into array Press_Data[], in hPa
    Timer_ReadAll(); // reads pulse timer into Timer_Data, in RPM ??? XXX
    UpdateCalibratedPressure();
    DoPressure();
    //DoFlow();
    DoSerialIn();
    DoEjector();
    DoServos();
    if (millis() >= nextTime2) {
      //MeasureElectricalPower();
      //accumulateEnergyValues();
      if (millis() >= nextTime1) {
        nextTime1 += loopPeriod1;
        //averageEnergyValues();
        //DoHertz();
        DoGrate();
        DoDatalogging();
        DoAlarmUpdate();
        grateOn = false;
        if (alarm == true) {
          analogWrite(ALARM_FET, 255);
        } else {
          analogWrite(ALARM_FET,0);
        }
        if (millis() >= nextTime0) {
          nextTime0 += loopPeriod0;
          DoAlarm();
        }
      }
    }
    // END USER CONTROL CODE
    UI_DoScr();       // output the display screen data, 
    // (default User Interface functions are in library KS/ui.c)
    // XXX should be migrated out of library layer, up to sketch layer                      
     key = Kpd_GetKeyAsync();
    // get key asynnchronous (doesn't wait for a keypress)
    // returns -1 if no key

    UI_HandleKey(key);  // the other two thirds of the UI routines:
    // given the key press (if any), then update the internal
    //   User Interface data structures
    // ALSO: Manipulate the various output data structures
    //   based on the keypad input

    Fet_WriteAll();   // Write the FET output data to the PWM hardware
    Servo_WriteAll(); // Write the Futaba hobby servo data to the PWM hardware

    PORTJ ^= 0x80;    // toggle the heartbeat LED
  }
}


void DoAlarmUpdate() {
  if (analogRead(54) > 256) {
    // auger on
    auger_on_length++;
    auger_off_length = max(0,auger_off_length*0.8-10);
  } else {
    // auger off
    auger_off_length++;
    auger_on_length = max(0,auger_on_length*.8-10);
  }
}

void DoAlarm() {
  alarm = false;
  if (auger_on_length >= auger_on_alarm_point) {
    Serial.println("# Auger on too long");
    alarm = true;
  }
  if (auger_off_length >= auger_off_alarm_point) {
    Serial.println("# Auger off too long");
    alarm = true;
  }
  if (Temp_Data[LOW_FUEL_TC] > 230) {
    Serial.println("# Reactor fuel may be low");
    alarm = true;
  }
}

// Datalogging
void LogTime(boolean header = false) {
  if (header) {
    Serial.print("#Time [desisec],");
  } else {
    Serial.print(millis()/100); // time since restart in deciseconds
    Serial.print(", ");
  }
}

void LogFlows(boolean header = false) {
  if (header) {
    Serial.print("Q_air_rct,Q_gas,");
  } else {
    Serial.print(air_rct_flow);
    Serial.print(", ");
    Serial.print(gas_flow);
    Serial.print(", ");
  }
}

void LogPID(boolean header = false) {
  if (header) {
    Serial.print("Ejector_In,Ejector_Out,Ejector_Setpoint,Ejector_P,Ejector_I,Ejector_D,");
  } else {
    Serial.print(ejector_input);
    Serial.print(", ");
    Serial.print(ejector_output);
    Serial.print(", ");
    Serial.print(ejector_setpoint);
    Serial.print(", ");
    Serial.print(ejector_PID.GetP_Param());
    Serial.print(", ");
    Serial.print(ejector_PID.GetI_Param());
    Serial.print(", ");
    Serial.print(ejector_PID.GetD_Param());
    Serial.print(", ");
  }
}

void LogAnalogInputs(boolean header = false) {
  if (header) {
    Serial.print("ANA0,ANA1,ANA2,ANA3,");
  } else {
    Serial.print(ADC_ReadChanSync(0));
    Serial.print(", ");  
    Serial.print(ADC_ReadChanSync(1));
    Serial.print(", ");
    Serial.print(ADC_ReadChanSync(2));
    Serial.print(", ");
    Serial.print(ADC_ReadChanSync(3));
    Serial.print(", ");
  }
}

void LogGrate(boolean header = false) {
  if (header) {
    Serial.print("Grate,P_ratio,P_ratio_state,");
  } else {
    if (grateOn) {
      Serial.print("grate_on");
    } else {
      Serial.print("grate_off");
    }
    Serial.print(", ");
    Serial.print(pRatio);
    Serial.print(", ");
    //Serial.print(P_comb);
    //Serial.print(", ");
    if (pRatioHigh) {
      Serial.print("pRatioHigh");
    } else {
      Serial.print("pRatioLow");
    }
    Serial.print(", ");
  }
}

void LogPressures(boolean header = false) {
  if (header) {
    if (GCU_fill == FULLFILL) {
      Serial.print("P0,P1,P2,P3,P4,P5,");
    } else {
      Serial.print("P0,P4,");
    }
  } else {
    if (GCU_fill == FULLFILL) {
      Serial.print(Press[0]);
      Serial.print(", ");
      Serial.print(Press[1]);
      Serial.print(", ");
      Serial.print(Press[2]);
      Serial.print(", ");
      Serial.print(Press[3]);
      Serial.print(", ");
      Serial.print(Press[4]);
      Serial.print(", ");
      Serial.print(Press[5]);
      Serial.print(", ");
    } else {
      Serial.print(Press[0]);
      Serial.print(", ");
      Serial.print(Press[4]);
      Serial.print(", ");
    }
  }
}

void LogTemps(boolean header = false) {
  if (header) {
    Serial.print("T0,T1,T2,T3,T4,T5,");
  } else {
    Serial.print(Temp_Data[0]);
    Serial.print(", ");
    Serial.print(Temp_Data[1]);
    Serial.print(", ");
    Serial.print(Temp_Data[2]);
    Serial.print(", ");
    Serial.print(Temp_Data[3]);
    Serial.print(", ");
    Serial.print(Temp_Data[4]);
    Serial.print(", ");
    Serial.print(Temp_Data[5]);
    Serial.print(", ");
    Serial.print(Temp_Data[6]);
    Serial.print(", ");
  }
}

void LogAuger(boolean header = false) {
  if (header) {
    Serial.print("Auger,");
  } else {
    Serial.print(analogRead(54)); //Phidgets "2" - auger sense on APL skid
    Serial.print(", ");
  }
}

void DoDatalogging() {
  boolean header = false;
  Serial.begin(57600); //reset serial?
  if (lineCount == 0) {
    header = true;
  }
  LogTime(header);
  LogTemps(header);
  LogPressures(header);
  //LogFlows(header);
  LogGrate(header);
  LogAnalogInputs(header);
  LogPID(header);
  LogAuger(header);
  Serial.println();
  lineCount++;
}

// Ejector
void Ejector_Init() {
  // ejector initialization
  ejector_setpoint = -1000;
  ejector_output = 0.5;
  ejector_PID.SetSampleTime(20);
  ejector_PID.SetInputLimits(-10000,0);
  ejector_PID.SetOutputLimits(0,1);
  ejector_PID.SetMode(AUTO);
  ejector_input = P_reactor;
  ejector_PID.Compute();
  ejector_nextCompute = millis() + ejector_period;
}

void DoEjector() {
  if (millis() >= ejector_nextCompute) {
    ejector_PID.SetTunings(ejector_P, ejector_I, ejector_D);
    ejector_input = P_reactor;
    ejector_PID.Compute();
    ejector_nextCompute += ejector_period;
  }
  if (millis() % ejector_period < ejector_output*1000) {
    digitalWrite(EJECTOR_FET, HIGH);
  } else {
    digitalWrite(EJECTOR_FET, LOW);
  }
}

// Flow

void DoFlow() {
  air_rct_flow = CalcFlowmeter(CfA0_air_rct, -Press[4], 1.2);
  gas_flow = CalcFlowmeter(CfA0_gas, -Press[5], 0.95);
}
//Correct flowmeter equation
double CalcFlowmeter(double CfA0, double dP, double density) {
  return CfA0*sqrt(((dP)*2)/density);
}


void DoGrate() {
  pRatio = (float)P_comb/(float)P_reactor;
  pRatioHigh = (pRatio < 0.4 && P_reactor < -200 && P_comb < -50);
  //if (millis() >= nextGrate-gratePulseLength) {
    if (grateMode == GRATE_ON || (pRatioHigh && grateMode == GRATE_PRATIO)) {
      analogWrite(GRATE_SOLENOID, 255);
      grateOn = true;
      //re-do brackets for period based shake
    } else {
      analogWrite(GRATE_SOLENOID,0);
      grateOn = false;
  }
  if (millis() >= nextGrate) {
    nextGrate += gratePeriod;
  }
}

// Pressure

void DoPressure() {
  P_reactor = Press[0];
  P_comb = Press[4];
  //P_comb_smooth = smooth(P_comb,0.999,P_comb_smooth);
}

void CalibratePressureSensors() {
  int P_sum[6] = {0,0,0,0};
  int P_ave;
  byte lowbyte,highbyte;
  Serial.println("Calibrating Pressure Sensors");
  for (int i=0; i<10; i++) {
    Press_ReadAll();
    for (int j=0; j<6; j++) {
      P_sum[j] += Press_Data[j];
    }
    delay(1);
  }
  //write to EEPROM
  for (int i=0; i<6; i++) {
    P_ave = float(P_sum[i])/10.0;
    lowbyte = ((P_ave >> 0) & 0xFF);
    highbyte = ((P_ave >> 8) & 0xFF);
    EEPROM.write(i*2, lowbyte);
    EEPROM.write(i*2+1, highbyte);
  }
}

void LoadPressureSensorCalibration() {
  int calib;
  byte lowbyte,highbyte;
  Serial.println("Loading Pressure Sensor Calibrations:");
  for (int i=0; i<6; i++) {
    byte lowByte = EEPROM.read(i*2);
    byte highByte = EEPROM.read(i*2 + 1);
    Press_Calib[i] = ((lowByte << 0) & 0xFF) + ((highByte << 8) & 0xFF00);
    Serial.print("P");
    Serial.print(i);
    Serial.print(": ");
    Serial.print(Press_Calib[i]);
    Serial.println();
  }
}

void UpdateCalibratedPressure() {
  for (int i = 0; i<6; i++) {
    Press[i] = Press_Data[i]-Press_Calib[i];
  }
}

float smooth(int data, float filterVal, float smoothedVal) {
  if (filterVal > 1){      // check to make sure param's are within range
    filterVal = .99;
  }
  else if (filterVal <= 0){
    filterVal = 0;
  }
  smoothedVal = (data * (1 - filterVal)) + (smoothedVal  *  filterVal);
  return smoothedVal;
}


// SerialIn
void DoSerialIn() {
  int incomingByte = 0;
  double p,i,d;
  double p_d =0.02;
  double i_d = 0.02;
  double d_d = 0.02;
  // Serial input
  if (Serial.available() > 0) {
    p=ejector_P;
    i=ejector_I;
    d=ejector_D;
    switch (Serial.read()) {
    case 'p':
      PrintEjectorUpdate(p,i,d,p+p_d,i,d);
      p=p+p_d;
      ejector_P=p;
      break;
    case 'P':
      PrintEjectorUpdate(p,i,d,p-p_d,i,d);
      p=p-p_d;
      ejector_P=p;
      break;
    case 'i':
      PrintEjectorUpdate(p,i,d,p,i+i_d,d);
      i=i+i_d;
      ejector_I=i;
      break;
    case 'I':
      PrintEjectorUpdate(p,i,d,p,i-i_d,d);
      i=i-i_d;
      ejector_I=i;
      break;
    case 'd':
      PrintEjectorUpdate(p,i,d,p,i,d+d_d);
      d=d+d_d;
      ejector_D=d;
      break;
    case 'D':
      PrintEjectorUpdate(p,i,d,p,i,d-d_d);
      d=d-d_d;
      ejector_D=d;
      break;
    case 'c':
      CalibratePressureSensors();
      LoadPressureSensorCalibration();
      break;
    case 's':
      servo1_pos += 10;
      Serial.print("#Servo1 (degrees) now:");
      Serial.println(servo1_pos);
      break;
    case 'S':
      servo1_pos -= 10;
      Serial.print("#Servo1 (degrees) now:");
      Serial.println(servo1_pos);
      break;
    case 'e':
      ejector_setpoint += 1000;
      Serial.print("#Ejector Setpoint now:");
      Serial.println(ejector_setpoint);
      break;
    case 'E':
      ejector_setpoint -= 1000;
      Serial.print("#Ejector Setpoint now:");
      Serial.println(ejector_setpoint);
      break;
    case 't':
      loopPeriod1 = min(loopPeriod1+100,loopPeriod2);
      Serial.print("#Sample Period now:");
      Serial.println(loopPeriod1);
      break;
    case 'T':
      loopPeriod1 = min(loopPeriod1-100,loopPeriod2);
      Serial.print("#Sample Period now:");
      Serial.println(loopPeriod1);
      break;
    case 'g':  
      analogWrite(GRATE_SOLENOID, 255);
      delay(gratePulseLength);
      analogWrite(GRATE_SOLENOID, 0);
      Serial.println("#Grate Shaken");
      grateOn = true;
      break;
    case 'G':  
      switch (grateMode) {
      case GRATE_OFF:
        grateMode = GRATE_ON;
        Serial.println("#Grate Mode: On");
        break;
      case GRATE_ON:
        grateMode = GRATE_PRATIO;
        Serial.println("#Grate Mode: Pressure Ratio");
        break;
      case GRATE_PRATIO:
        grateMode = GRATE_OFF;
        Serial.println("#Grate Mode: Off");
        break;
      }
      break;  
    case 'm':
      gratePeriod += 500;
      Serial.print("#Grate Interval now:");
      Serial.println(gratePeriod);
      break;
    case 'M':
      gratePeriod -= 500;
      Serial.print("#Grate Interval now:");
      Serial.println(gratePeriod);
      break;   
    }
  }
}

void PrintEjectorUpdate(double P, double I, double D, double nP, double nI, double nD) {
  Serial.print("#Updating ejector PID from [");
  Serial.print(P);
  Serial.print(",");
  Serial.print(I);
  Serial.print(",");
  Serial.print(D);
  Serial.print("] to [");
  Serial.print(nP);
  Serial.print(",");
  Serial.print(nI);
  Serial.print(",");
  Serial.print(nD);
  Serial.println("]");
}



// Servo
void PulseServo(int servo_pin,double angle) {
  digitalWrite(servo_pin,HIGH);
  delayMicroseconds(1520+angle*6);
  digitalWrite(servo_pin,LOW);
}

// code for driving servos w/ dead band
void DoServos() {
  if (servo_alt == 1) { // pulse every other time through the loop
    if (abs(servo0_pos - servo0_db) > 3) {
      servo0_db = servo0_pos;
    }
    if (abs(servo1_pos - servo1_db) > 3) {
      servo1_db = servo1_pos;
    }
    PulseServo(SERVO0,servo0_db);
    PulseServo(SERVO1,servo1_db);
    servo_alt = 0; 
  }
  servo_alt++;
}

int main(void)
{
	init();

	setup();
    
	for (;;)
		loop();
        
	return 0;
}

